import 'dart:async';

import 'package:fis_common/event/event_type.dart';
import 'package:web_socket_channel/web_socket_channel.dart';
import 'package:web_socket_channel/status.dart' as statuses;

import 'core/interface/connection.dart';

class WsConnectionException implements Exception {
  final dynamic message;

  WsConnectionException(this.message);
}

/// WebScoket连接
class WsConnection implements IConnection {
  /// 错误断开重试次数
  static const C_ERR_RETRY_LIMIT = 5;

  /// 连接请求超时时间，10s
  static const C_CONNECT_REQ_TIMEOUT = 10 * 1000;

  ConnectionStatus _status = ConnectionStatus.init;

  int _errorRetryCount = 0;

  Timer? _timer;

  /// 主机地址
  final String host;

  WebSocketChannel? _channel;

  WsConnection(
    this.host, {
    this.exceptionOccurred,
    this.statusChanged,
  }) {
    messageReceived = FEventHandler<dynamic>();
  }

  /// 连接状态
  ConnectionStatus get status => _status;

  /// 是否已连接
  bool get isConnected => status == ConnectionStatus.connected;

  @override
  FEventHandler<Exception>? exceptionOccurred;

  @override
  late final FEventHandler<dynamic> messageReceived;

  @override
  FEventHandler<ConnectionStatus>? statusChanged;

  @override
  Future<bool> connect() {
    final completer = Completer<bool>();
    final uri = Uri.parse(host);
    _channel = WebSocketChannel.connect(uri);
    _updateStatus(ConnectionStatus.connecting);

    _channel!.stream.listen(
      (data) {
        messageReceived.emit(this, data);
      },
      onError: (error) {
        _handleCancelOnError();
        exceptionOccurred?.emit(this, WsConnectionException(error));
      },
      cancelOnError: true,
    );
    _waitConnectResult(completer);
    return completer.future;
  }

  @override
  Future<bool> close() async {
    _updateStatus(ConnectionStatus.disconnected);
    if (_channel != null) {
      if (_channel!.closeCode == null) {
        await _channel!.sink.close(statuses.normalClosure);
      }
      _channel = null;
    }
    _timer?.cancel();
    _timer = null;
    return true;
  }

  Future<bool> _checkIsConnected() async {
    if (_channel == null) {
      return false;
    } else {
      final streamDone = await _channel!.sink.done;
      return streamDone == false;
    }
  }

  void _updateStatus(ConnectionStatus value) {
    if (value != _status) {
      _status = value;
      statusChanged?.emit(this, value);
    }
  }

  /// 等待请求连接结果
  void _waitConnectResult(Completer completer) {
    const onceDuration = 100;
    const limit = C_CONNECT_REQ_TIMEOUT / onceDuration;
    int count = 0;
    _timer = Timer.periodic(
      const Duration(milliseconds: onceDuration),
      (timer) async {
        if (count >= limit) {
          // 等待连接超时
          _updateStatus(ConnectionStatus.disconnected);
          exceptionOccurred?.emit(
              this, WsConnectionException("WsConnectionTimeout"));
          completer.complete(false);
          timer.cancel();
          _timer = null;
          return;
        }

        final result = await _checkIsConnected();
        if (result) {
          if (!completer.isCompleted) {
            _updateStatus(ConnectionStatus.connected);
            completer.complete(true);
            timer.cancel();
            _timer = null;
          }
        }
      },
    );
    //C_CONNECT_REQ_TIMEOUT
  }

  /// 处理因错误导致的连接中断
  void _handleCancelOnError() {
    if (!isConnected) return; // 非连接状态，不处理

    _updateStatus(ConnectionStatus.disconnected);

    if (_errorRetryCount >= C_ERR_RETRY_LIMIT) {
      // 超过重试次数，不处理
      exceptionOccurred?.emit(
          this, WsConnectionException("OutOfErrorRetryLimit"));
      return;
    }
    _errorRetryCount++;
    connect();
  }
}
